#include <map>

#include "i3d/image3d.h"
#include "mylist.h"
#include "histogram.h"

using namespace i3d;

void ComputeHistogram(const Image3d<GRAY8>& img, int (&histogram)[256])
{

	//initiate histogram to zeros everywhere
	for (int i=0; i < 256; ++i)
		histogram[i]=0;

	//sweep the image and count how many times each pixel values was used
	const GRAY8 *p=img.GetFirstVoxelAddr();
	for (size_t i=0; i < img.GetImageSize(); ++i, ++p)
		++histogram[*p];
}


// ---------------- mylist ---------------- 
void ComputeHistogram_MyList(const i3d::Image3d<i3d::GRAY8>& img)
{
	struct Elem *h=NULL;

	//pocet celkem provedenych porovnani
	int counter=0;

	//vytvori histogram
	const GRAY8 *p=img.GetFirstVoxelAddr();
	for (size_t i=0; i < img.GetImageSize(); ++i, ++p)
		counter += AddToHistogram(&h,*p);

/*
	//vytiskne histogram na obrazovku
	while (h != NULL)
	{
		std::cout << "hodnota " << h->intensity << " je v obraze "
			<< h->count << " krat\n";
		h=h->next;
	}
 */

	std::cout << "MyList   : pro vypocet bylo potreba provest " << counter << " porovnani\n";

	//tady by se jeste mela uvolnit alokovana pamet!
}


void ComputeHistogram_MyListOpt(const i3d::Image3d<i3d::GRAY8>& img)
{
	struct Elem *h=NULL,*last=NULL;
	//h je samotny histogram
	//last je ukazatel na posledne pridavany prvek

	//pocet celkem provedenych porovnani
	int counter=0;

	//vytvori histogram
	const GRAY8 *p=img.GetFirstVoxelAddr();
	for (size_t i=0; i < img.GetImageSize(); ++i, ++p)
		counter += AddToHistogram(&h,&last,*p);

	/*
	//vytiskne histogram na obrazovku
	while (h != NULL)
	{
		std::cout << "hodnota " << h->intensity << " je v obraze "
			<< h->count << " krat\n";
		h=h->next;
	}
	 */

	std::cout << "MyListOpt: pro vypocet bylo potreba provest " << counter << " porovnani\n";

	//tady by se jeste mela uvolnit alokovana pamet!
}


// ---------------- std::map ---------------- 
/**
 * Global counter of std::map effectivity. It tells how many times comparison happened 
 * when a map is accessed. The counter doesn't however distinguish between insertions or
 * plain reads of the map. Hence, interpret its value with caution.
 */
int CountCompares=0;

/**
 * A comparison class for std::map. It, hopefully, provides the same service as the original
 * implicit comparison class in the std::map. On top of it, it increments the global counter
 * CountCompares everytime it is "executed".
 */
class Compare {
  public:
	bool operator() (const int lhs,const int rhs) {
	  //for studying purposes...
	  //cout << "comparing " << lhs << " with " << rhs << "\n";

	  ++CountCompares;
	  return(lhs < rhs);
	}
};


void ComputeHistogram_StdMap(const i3d::Image3d<i3d::GRAY8>& img)
{
	//the sparse histogram
	std::map<int,int,Compare> hist;

	//sweep the image and count how many times each pixel values was used
	const GRAY8 *p=img.GetFirstVoxelAddr();
	for (size_t i=0; i < img.GetImageSize(); ++i, ++p)
		++hist[*p];

	std::cout << "std::map : pro vypocet bylo potreba provest " << CountCompares << " porovnani\n";
}
