#include <iostream>
#include "i3d/image3d.h"

//time measurement - LINUX_SPECIFIC
#include <sys/time.h>


using namespace std;
using namespace i3d;

void ThresholdPOINTERS(Image3d<GRAY8> &img,const GRAY8 thres) {

	GRAY8 *p=img.GetFirstVoxelAddr();
	for (size_t i=0; i < img.GetImageSize(); ++i, ++p) {
		*p=(*p > thres)? 1 : 0;
	}
}

void ThresholdINDICES(Image3d<GRAY8> &img,const GRAY8 thres) {

	for (size_t i=0; i < img.GetImageSize(); ++i) {
		img.SetVoxel(i,
				(img.GetVoxel(i) > thres)? 1 : 0
				    );
	}
}

void ThresholdFORCYCLES(Image3d<GRAY8> &img,const GRAY8 thres) {

	for (size_t z=0; z < img.GetSizeZ(); ++z)
	  for (size_t y=0; y < img.GetSizeY(); ++y)
	    for (size_t x=0; x < img.GetSizeX(); ++x)
	    	{
			img.SetVoxel( x,y,z,
				(img.GetVoxel(x,y,z) > thres)? 1 : 0
				    );
		}
}

void ThresholdFORCYCLES_BAD(Image3d<GRAY8> &img,const GRAY8 thres) {

	for (size_t x=0; x < img.GetSizeX(); ++x)
	  for (size_t y=0; y < img.GetSizeY(); ++y)
	    for (size_t z=0; z < img.GetSizeZ(); ++z)
	    	{
			img.SetVoxel( x,y,z,
				(img.GetVoxel(x,y,z) > thres)? 1 : 0
				    );
		}
}

int main(void) {
	//opens and reads GRAY 8bit input image
	Image3d<GRAY8> img;
	img.ReadImageICS("images/cell04_3D_GRAY8.ics");

	//dry run - to fill caches
	ThresholdPOINTERS(img,85);
	ThresholdPOINTERS(img,85);

	//time-measurement: time stamp before - LINUX_SPECIFIC
	struct timeval lt1,lt2,lt3,lt4,lt5;
	gettimeofday(&lt1,NULL);

	//the measured routine
	ThresholdPOINTERS(img,85);

	//time-measurement: time stamp after
	//...and before for the next measurement at the same time
	// - LINUX_SPECIFIC
	gettimeofday(&lt2,NULL);

	ThresholdINDICES(img,85);

	gettimeofday(&lt3,NULL);

	ThresholdFORCYCLES(img,85);

	gettimeofday(&lt4,NULL);

	ThresholdFORCYCLES_BAD(img,85);

	gettimeofday(&lt5,NULL);

	//remove the common higher orders from the numbers
	//to avoid subtracting numbers of different magnitudes,
	//which is something floating-point arithmetic does not like
	lt2.tv_sec-=lt1.tv_sec;
	lt3.tv_sec-=lt1.tv_sec;
	lt4.tv_sec-=lt1.tv_sec;
	lt5.tv_sec-=lt1.tv_sec;

	//compute the time required using seconds and microseconds
	float T1,T2;
	T1=(float)lt1.tv_usec / 1000000.0f;
	T2=(float)lt2.tv_sec + ((float)lt2.tv_usec / 1000000.0f);
	cout << "ThresholdPOINTERS() needed approx. " << T2-T1 << " seconds\n";

	T1=(float)lt3.tv_sec + ((float)lt3.tv_usec / 1000000.0f);
	cout << "ThresholdINDICES() needed approx. " << T1-T2 << " seconds\n";

	T2=(float)lt4.tv_sec + ((float)lt4.tv_usec / 1000000.0f);
	cout << "ThresholdFORCYCLES() needed approx. " << T2-T1 << " seconds\n";

	T1=(float)lt5.tv_sec + ((float)lt5.tv_usec / 1000000.0f);
	cout << "ThresholdFORCYCLES_BAD() needed approx. " << T1-T2 << " seconds\n";

	return(0);
}
